-----------------------------------------------------------------------------
--
-- data
--  \threat
--    \<playerName>
--      - amount
--		\targetDetails
--        <targetHash> : amount
--
-----------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("AssessmentThreat")
L:RegisterTranslations("enUS", function() return {
	["Threat List"] = true,
	["Threat Deficit"] = true,
	["Aggro Gain"] = true,
	["Aggro Gain Color"] = true,
	["Threat"] = true,
	["Threat gap before aggro gain on target"] = true,

	["Single Target Mode"] = true,
	["Single target mode causes Assessment to clear threat on mobs of that name when they die."] = true,

	["Threat"] = true,
	["Threat Options."] = true,
	["Aggro Gain Color"] = true,
	["Aggro gain color"] = true,
	
	["Group Threat Versions"] = true,

	["Aggro Gain Text"] = true,
	["Text to show for the aggro gain bar."] = true,
	["Any Text"] = true,

	["Enable KLHTM Broadcast"] = true,
	["Broadcast threat data to KLHThreatMeter users."] = true,
} end)

L:RegisterTranslations("koKR", function() return {
	["Threat List"] = "위협 목록",
	["Threat Deficit"] = "위협 결손치",
	["Aggro Gain"] = "어그로 획득",
	["Aggro Gain Color"] = "어그로 획득 색상",
	["Threat"] = "위협",
	["Threat gap before aggro gain on target"] = "대상에 어그로 획득 이전 위협 차이",

	["Single Target Mode"] = "단순 대상 모드",
	["Single target mode causes Assessment to clear threat on mobs of that name when they die."] = "단순 대상 모드는 같은 이름의 몹이 죽었을 때 명확한 위협을 평가합니다.",

	["Threat"] = "위협",
	["Threat Options."] = "위협 설정입니다.",
	["Aggro Gain Color"] = "어그로 획득 색상",
	["Aggro gain color"] = "어그로 획득 색상입니다.",
	
	["Group Threat Versions"] = "파티 위협 버전",

	["Aggro Gain Text"] = "어그로 획득 글자",
	["Text to show for the aggro gain bar."] = "어그로 획득 바에 표시할 글자입니다.",
	["Any Text"] = "문자열",

	["Enable KLHTM Broadcast"] = "KLHTM 알림 사용",
	["Broadcast threat data to KLHThreatMeter users."] = "KLHThreatMeter 사용자에게 위협 데이터를 알립니다.",
} end)

local AssessmentThreat = Assessment:NewModule("Threat", "AceEvent-2.0", "Parser-3.0")
AssessmentThreat.db = Assessment:AcquireDBNamespace("Threat")
Assessment:RegisterDefaults("Threat", "profile", {
	enable				= true,
	singleTargetMode	= false,
	aggroGainColor		= { 1, 0, 0 },
	aggroGainText		= L["Aggro Gain"],
	klhtmBroadcast		= false,
})

local Threat = AceLibrary("Threat-1.0")
local THREAT_GLOBAL_TARGET = Threat.GlobalTarget

-----------------------------------------------------------------------------

local aggroGainText
local profile

local emptyMethod = Assessment.EmptyMethod
local new, del = Assessment.NewTable, Assessment.ReleaseTable

-----------------------------------------------------------------------------

local function GetAggroGainThreatMultiplier()
	return Threat:UnitInMeleeRange("target") and 1.1 or 1.3
end

local function GetAggroGainThreat(targetName)
	if UnitIsUnit("player", "targettarget") then
		return Threat:GetThreat(UnitName("player"), targetName)
	end
	
	local multiplier = GetAggroGainThreatMultiplier()
	local targetTargetName = UnitName("targettarget")
	if not targetTargetName then
		return Threat:GetMaxThreatOnTarget(targetName)
	end
	return Threat:GetThreat(targetTargetName, targetName) * multiplier
end

-----------------------------------------------------------------------------

Assessment.TagData.TargetThreat =
{
	method = function(data)
				local targetName = UnitName("target")
				if not targetName then return 0 end
				if data.key == aggroGainText then
					return GetAggroGainThreat(targetName)
				else
					return Threat:GetThreat(data.key, targetName)
				end
			 end,
	data			= { "target", "threat" },
	guaranteeResult = true,
	numeric			= true
}

Assessment.TagData.MaxTargetThreat =
{
	method = function(data)
				local targetName = UnitName("target")
				if not targetName then return 0 end
				return Threat:GetMaxThreatOnTarget(targetName)
			 end,
	data			= { "target", "threat" },
	guaranteeResult = true,
	numeric			= true
}

Assessment.TagData.AggroGainTargetThreat =
{
	method = function(data)
				local targetName = UnitName("target")
				if not targetName then return 0 end

				return GetAggroGainThreat(targetName)
			 end,
	data			= { "target", "threat" },
	guaranteeResult = true,
	numeric			= true
}

Assessment.TagData.Threat =
{
	method = function(data)
				if data.key == aggroGainText then return 0 end

				local fightData = data.fightData
				if not fightData then return 0 end
				
				local threat = fightData.threat
				if not threat then return 0 end
				
				local playerThreat = threat[data.key]
				if not playerThreat then return 0 end

				return playerThreat.amount					
			 end,
	data			= { "threat" },
	guaranteeResult = true,
	numeric			= true
}

Assessment.TagData.ThreatVersion =
{
	method = function(data, prior)
				if not prior then return end
				if data.key == UnitName("player") then
					local _, v = Threat:GetLibraryVersion()
					return v
				else
					return Threat:GetPartyRevisions()[data.key]
				end
			 end,
}

Assessment.TagData.MaxThreat =
{
	method = function(data)
				local fightData = data.fightData
				if not fightData then return 0 end
				
				local threat = fightData.threat
				if not threat then return 0 end

				local MaxThreat = 0
				for k,v in pairs(threat) do
					if v.amount > MaxThreat then
						MaxThreat = v.amount
					end
				end
				return MaxThreat
			 end,
	data			= { "threat" },
	guaranteeResult = true,
	numeric			= true
}

Assessment.TagData.ThreatTargetName =
{
	method = function(data) 
		if UnitExists("target") and not UnitIsFriend("player", "target") then
			return UnitName("target")
		end
	end,
}

Assessment.TagData.IsUsingForeignThreatSource =
{
	method = function(data) return Threat:IsUsingForeignThreatSource(data.key) end,
}

-----------------------------------------------------------------------------

local function ThreatListIterator(state, index)
	if not index then
		return aggroGainText
	elseif index == aggroGainText then
		return next(state)
	else
		return next(state, index)
	end
end

Assessment.ListType.ThreatList =
{
    method = function(fightData)
                if not fightData.threat then return emptyMethod end
				if UnitExists("targettarget") and not UnitIsFriend("player", "target") and not UnitIsUnit("player", "targettarget") then
					return ThreatListIterator, fightData.threat, nil
				end
				return pairs(fightData.threat)
             end,
    data = "threat",
    description = L["Threat List"],
}

-----------------------------------------------------------------------------

Assessment.ColorType.AggroGain =
{
	method = function(fightData, data)
				if data.key ~= aggroGainText then return nil end
				return profile.aggroGainColor
			 end,
	description = L["Aggro Gain Color"]
}

-----------------------------------------------------------------------------

Assessment.DefaultDisplaySets.Threat =
{
	titleTag		= "[ThreatTargetName:Yellow||\"Threat\"]",
	listType		= "ThreatList", 
	barValueTag		= "Select(ThreatTargetName, TargetThreat, Threat)", 
	filterTag		= "BarValue ~= 0",
	fightIndex		= 0,
	text			= { 
						{ tag = "[Rank]. [Key]", offsets = { right = -45 } }, 
						{ tag = "[IsUsingForeignThreatSource:\"*\"][BarValue:Short]" },
						{ tag = "[(Select(ThreatTargetName, TargetThreat, Threat)/Maximum(CurrentCombatDuration, 1)):Format(\"%.1f\")]", alignment = "RIGHT", offsets = { right = -25 } },
						{ tag = "[(Select(ThreatTargetName, TargetThreat/MaxTargetThreat, Threat/MaxThreat)*100):Percent]", alignment = "RIGHT", visible = false, offsets = { right = -50 }  },
						count = 4,
					  },
	description		= L["Threat"],
	colorTypeList	= { "AggroGain", "Player", "Class", count = 3 },
}

Assessment.DefaultDisplaySets.GroupThreatVersions =
{
	titleTag		= "Threat Versions [ThreatVersion:Prefix(\"r\"):Yellow]",
	listType		= "CurrentGroupMembers", 
	barValueTag		= "ThreatVersion",
	sortKeyTag		= "BarValue",
	filterTag		= "1",
	fightIndex		= 0,
	text			= { { tag = "[Key]" }, { tag= "[select(ThreatVersion, ThreatVersion:Yellow, \"Unknown\":Red)]",  } },
	description		= L["Group Threat Versions"],
	defaultColor	= { 0.3, 0.5, 0.8 },
	colorTypeList	= { "Class", count = 1 },
	favorite		= false,
}

-----------------------------------------------------------------------------

Assessment.DefaultTags.ThreatDeficit =
{
	tag			= "[(MaxTargetThreat-TargetThreat):Short]",
	fightIndex	= 0,
	description	= L["Threat Deficit"]
}

Assessment.DefaultTags.TargetAggroThreatDeficit =
{
	tag			= "[(AggroGainTargetThreat-TargetThreat):Short]",
	fightIndex	= 0,
	description	= L["Threat gap before aggro gain on target"]
}

-----------------------------------------------------------------------------

Assessment.Options.args.data.args.threat =
{
	type = "group",
	name = L["Threat"],
	desc = L["Threat Options."],
	args =
	{
		aggroGainColor =
		{
			type = "color",
			name = L["Aggro Gain Color"],
			desc = L["Aggro gain color"],
			get = function() return unpack(profile.aggroGainColor) end,
			set = function(r, g, b) profile.aggroGainColor = { r, g, b, a } Assessment:RedrawAll() end,
		},
		singleTargetMode =
		{
			type = "toggle",
			name = L["Single Target Mode"],
			desc = L["Single target mode causes Assessment to clear threat on mobs of that name when they die."],
			get = function() return profile.singleTargetMode end,
			set = function(v) AssessmentThreat:SetSingleTargetMode(v) end,
		},
		aggroGainText =
		{
			type = "text",
			name = L["Aggro Gain Text"],
			desc = L["Text to show for the aggro gain bar."],
			get = function() return profile.aggroGainText end,
			set = function(v) profile.aggroGainText = v aggroGainText = v Assessment:RedrawAll() end,
			usage = L["Any Text"],
		},
		klhtmBroadcast =
		{
			type = "toggle",
			name = L["Enable KLHTM Broadcast"],
			desc = L["Broadcast threat data to KLHThreatMeter users."],
			get = function() return Threat:IsKLHTMBroadcastEnabled() end,
			set = function(v) profile.klhtmBroadcast = v Threat:EnableKLHTMBroadcast(v) end,
		},
	}
}

-----------------------------------------------------------------------------

function AssessmentThreat:OnInitialize()
	profile = self.db.profile
	aggroGainText = profile.aggroGainText
end

function AssessmentThreat:OnEnable()
	self:RegisterEvent("PLAYER_TARGET_CHANGED", "OnPlayerTargetChanged")
	self:RegisterEvent("Threat_ThreatUpdated", "OnThreatUpdated")
	self:RegisterEvent("Threat_ThreatCleared", "OnThreatCleared")
	
	if profile.singleTargetMode then
		self:RegisterParserEvent({ eventType = "Death" }, "OnParserDeathEvent")
	end

	Threat:EnableKLHTMBroadcast(profile.klhtmBroadcast)
end

function AssessmentThreat:OnDisable()
end

-----------------------------------------------------------------------------

function AssessmentThreat:SetSingleTargetMode(v)
	profile.singleTargetMode = v
	
	if v then
		self:RegisterParserEvent({ eventType = "Death" }, "OnParserDeathEvent")
	else
		self:UnregisterAllParserEvents()
	end
end

function AssessmentThreat:OnPlayerTargetChanged()
	for data in self:ForceIterateUpdateDataTables() do
		data.dataUpdateTimes.target = GetTime()
	end
end

function AssessmentThreat:OnThreatCleared()
	local data = Assessment:GetFightData(1)
	if data and data.threat then
		data.threatLocked = true
	end
	
	data = Assessment:GetFightData(0)
	if data.threat then
		data.dataUpdateTimes.threat = GetTime()
		data.threat = del(data.threat, true)
	end
end

function AssessmentThreat:OnThreatUpdated(playerName, playerUnit, targetName)
	local threatValue = Threat:GetThreat(playerName, targetName)
	local time = GetTime()

	for data in self:ForceIterateUpdateDataTables() do
		data.dataUpdateTimes.threat = time
		
		local threat = data.threat
		if not threat then
			threat = new()
			data.threat = threat
		end
		
		local playerData = threat[playerName]
		if playerData then
			if targetName == THREAT_GLOBAL_TARGET and threatValue == 0 then
				if not data.threatLocked then
					threat[playerName] = del(playerData, true)
				end
			else
				if threatValue == 0 then
					playerData.targetDetails[targetName] = nil
				else
					playerData.targetDetails[targetName] = threatValue
				end
				
				if threatValue >= playerData.amount then
					playerData.amount = threatValue
				else
					local maximumThreatValue = threatValue
				
					for k,v in pairs(playerData.targetDetails) do
						if v > maximumThreatValue then
							maximumThreatValue = v
						end
					end
					
					if maximumThreatValue == 0 then
						threat[playerName] = del(playerData, true)
					else
						playerData.amount = maximumThreatValue
					end
				end
			end
		elseif threatValue ~= 0 then
			playerData = new()
			playerData.amount = threatValue
			playerData.targetDetails = new()
			playerData.targetDetails[targetName] = threatValue
			threat[playerName] = playerData
		end
	end
end

function AssessmentThreat:OnParserDeathEvent(eventData)
	-- Only clear threat on things killed by yourself, or someone in your group.
	if not eventData.sourceID or not UnitAffectingCombat("player") then return end

	Threat:GetPlayerModule():_setTargetThreat(0, eventData.recipientName)
	if UnitExists("pet") then
		Threat:GetPetModule():_setTargetThreat(0, eventData.recipientName)
	end
end

-----------------------------------------------------------------------------
